
import sys
from math import exp, log
#import networkx.algorithms.non_randomness

import sys
import logging

log = logging.getLogger('arg_dashboard.arg')
# handler = logging.StreamHandler(sys.stderr)
# handler.setFormatter(logging.Formatter('%(name)s - %(levelname)s - %(message)s'))
# log.addHandler(handler)
log.setLevel(logging.INFO)
import json

import random
# print("Setting random seed")
# random.seed(3)
# numpy.random.seed(3)

from random import random, shuffle, choices
from functools import partial
from copy import deepcopy, copy
# import networkx as nx
from functools import reduce
from numpy.random import exponential, sample
import numpy as np
from time import sleep
# import matplotlib.pyplot as plt
from collections import defaultdict


class Lineage(object):
    """
    Lineages are the edges of our graph
    """
    def __init__(self, lineageid=None, down=None, up=None, intervals=None):
        self.lineageid = lineageid
        self.down = down # node at bottom of edte
        self.up = up # node at top of edge
        self.intervals = intervals

    def __hash__(self):
        return self.lineageid

    def __eq__(self, other):
        return hasattr(other, 'lineageid') and self.lineageid == other.lineageid  

    def __repr__(self):
        return f'{self.lineageid}:{self.intervals}'

    def get_dict(self):
        d = self.__dict__.copy()
        if d['up']:
            d['up'] = d['up'].nodeid
        d['down'] = d['down'].nodeid
        return d

    def toJSON(self):
        return json.dumps(self, default=self.get_dict, sort_keys=True, indent=4)

    def __deepcopy__(self, memo):
        if self in memo:
            return memo.get(self)
        dup = type(self)(
            lineageid=deepcopy(self.lineageid, memo),
            # down=deepcopy(self.down, memo),
            # up=deepcopy(self.up, memo),
            intervals=deepcopy(self.intervals, memo)
            )
        memo[self] = dup
        dup.down=deepcopy(self.down, memo)
        dup.up=deepcopy(self.up, memo)
        return dup

class Node():
    """
    Leaf, Coalscent and Recombination are the nodes of the graph
    """
    def __hash__(self):
        return self.nodeid

    def __eq__(self, other):
        return hasattr(other, 'nodeid') and self.nodeid == other.nodeid  

    def __repr__(self):
        return f'{self.nodeid}'

class Leaf(Node):

    def __init__(self, nodeid=None, height=None, parent=None, intervals=[(0, 1)], xpos=None):
        self.nodeid = nodeid
        self.height = height
        self.intervals = intervals
        self.parent = parent
        self.xpos = xpos

    def get_dict(self):
        d = self.__dict__.copy()
        d['parent'] = d['parent'].lineageid
        return d

    def toJSON(self):
        return json.dumps(self, default=self.get_dict, sort_keys=True, indent=4)

    def __deepcopy__(self, memo):
        if self in memo:
            return memo.get(self)
        dup = type(self)(
            nodeid=deepcopy(self.nodeid, memo),
            height=deepcopy(self.height, memo),
            # parent=deepcopy(self.parent, memo),
            intervals=deepcopy(self.intervals, memo),
            xpos=deepcopy(self.xpos, memo)
            )
        memo[self] = dup
        dup.parent = deepcopy(self.parent, memo)
        return dup            

class Coalescent(Node):

    def __init__(self, nodeid=None, height=None, children=None, parent=None, xpos=None):
        self.nodeid = nodeid
        self.height = height
        self.children = children
        self.parent = parent
        self.xpos = xpos

    def get_dict(self):
        d = self.__dict__.copy()
        d['children'] = [c.lineageid for c in d['children']]
        d['parent'] = d['parent'].lineageid
        return d

    def toJSON(self):
        return json.dumps(self, default=self.get_dict, sort_keys=True, indent=4)

    def __deepcopy__(self, memo):
        if self in memo:
            return memo.get(self)
        dup = type(self)(
            nodeid=deepcopy(self.nodeid, memo),
            height=deepcopy(self.height, memo),
            # children=deepcopy(self.children, memo),
            # parent=deepcopy(self.parent, memo),
            xpos=deepcopy(self.xpos, memo)
            )
        memo[self] = dup
        dup.children=[deepcopy(c, memo) for c in self.children]
        dup.parent=deepcopy(self.parent, memo)
        return dup            

class Recombination(Node):

    def __init__(self, nodeid=None, height=None, child=None, recomb_point=None, left_parent=None, right_parent=None, xpos=None):
        self.nodeid = nodeid
        self.height = height
        self.child = child
        self.left_parent = left_parent
        self.right_parent = right_parent
        self.recomb_point = recomb_point
        self.xpos = xpos

    def get_dict(self):
        d = self.__dict__.copy()
        d['left_parent'] = d['left_parent'].lineageid
        d['right_parent'] = d['right_parent'].lineageid
        d['child'] = d['child'].lineageid
        return d

    def toJSON(self):
        return json.dumps(self, default=self.get_dict, sort_keys=True, indent=4)

    def __deepcopy__(self, memo):
        if self in memo:
            return memo.get(self)
        dup = type(self)(
            nodeid=deepcopy(self.nodeid, memo),
            height=deepcopy(self.height, memo),
            # child=deepcopy(self.child, memo),
            recomb_point=deepcopy(self.recomb_point, memo),
            # left_parent=deepcopy(self.left_parent, memo),
            # right_parent=deepcopy(self.right_parent, memo),
            xpos=deepcopy(self.xpos, memo)
            )  
        memo[self] = dup
        dup.child=deepcopy(self.child, memo)
        dup.left_parent=deepcopy(self.left_parent, memo)
        dup.right_parent=deepcopy(self.right_parent, memo)
        return dup                  

def flatten(list_of_tps):
 
    return reduce(lambda ls, ival: ls + list(ival), list_of_tps, [])


def unflatten(list_of_endpoints):

    return [ [list_of_endpoints[i], list_of_endpoints[i + 1]]
          for i in range(0, len(list_of_endpoints) - 1, 2)]

def merge(query, annot, op):

    a_endpoints = flatten(query)
    b_endpoints = flatten(annot)

    assert a_endpoints == sorted(a_endpoints), "not sorted or non-overlaping"
    assert b_endpoints == sorted(b_endpoints), "not sorted or non-overlaping"


    sentinel = max(a_endpoints[-1], b_endpoints[-1]) + 1
    a_endpoints += [sentinel]
    b_endpoints += [sentinel]

    a_index = 0
    b_index = 0

    res = []

    scan = min(a_endpoints[0], b_endpoints[0])
    while scan < sentinel:
        in_a = not ((scan < a_endpoints[a_index]) ^ (a_index % 2))
        in_b = not ((scan < b_endpoints[b_index]) ^ (b_index % 2))
        in_res = op(in_a, in_b)

        if in_res ^ (len(res) % 2):
            res += [scan]
        if scan == a_endpoints[a_index]: 
            a_index += 1
        if scan == b_endpoints[b_index]: 
            b_index += 1
        scan = min(a_endpoints[a_index], b_endpoints[b_index])

    return unflatten(res)

def interval_diff(a, b):
    if not (a and b):
        return a and a or b
    return merge(a, b, lambda in_a, in_b: in_a and not in_b)

def interval_union(a, b):
    if not (a and b):
        return []
    return merge(a, b, lambda in_a, in_b: in_a or in_b)

def interval_intersect(a, b):
    if not (a and b):
        return []
    return merge(a, b, lambda in_a, in_b: in_a and in_b)
    
def interval_sum(intervals):
    return sum(e - s for (s, e) in intervals)

def interval_span(intervals):
    starts, ends = zip(*intervals)
    return max(ends) - min(starts)

def interval_split(intervals, pos):
    left, right = list(), list()
    for s, e in intervals:
        if pos >= e:
            left.append((s, e))
        elif pos >= s and pos < e:
            left.append((s, pos))
            right.append((pos, e))
        else:
            right.append((s, e))
    return left, right

def interval_any_shared_borders(a, b):
    log.info('shared borders')
    flat = flatten(a) + flatten(b)
    return len(flat) > len(set(flat))

def x_positions_traverse(node, offset):
    """
    Recursive function for adding x positions
    """
    offset -= 1
    if type(node) is Leaf:
        return 1 # offset
        # do nothing
    if type(node) is Recombination:
        # add xpos to children
        node.child.down.xpos = node.xpos
        offset = x_positions_traverse(node.child.down, offset)
        # add xpos to node if both parents are positioned
        if node.left_parent.up.xpos is not None and node.right_parent.up.xpos is not None:
            left, right = sorted([node.left_parent.up.xpos, node.right_parent.up.xpos])
            x = left + (right - left) / 2 
            node.xpos = x
    if type(node) is Coalescent:
        # add xpos to children
        node.children[0].down.xpos = node.xpos - offset
        node.children[1].down.xpos = node.xpos + offset
        for child in node.children:
            x_positions_traverse(child.down, offset)

####################################################

def segments_crossing(line1, line2):
    xdiff = (line1[0][0] - line1[1][0], line2[0][0] - line2[1][0])
    ydiff = (line1[0][1] - line1[1][1], line2[0][1] - line2[1][1])

    def det(a, b):
        return a[0] * b[1] - a[1] * b[0]

    div = det(xdiff, ydiff)
    if div == 0:
        return False
    #    raise Exception('lines do not intersect')

    d = (det(*line1), det(*line2))
    x = det(d, xdiff) / div
    y = det(d, ydiff) / div

    # return x, y

    return \
        min(line1[0][0], line1[1][0]) < x < max(line1[0][0], line1[1][0]) and \
        min(line2[0][0], line2[1][0]) < x < max(line2[0][0], line2[1][0]) and \
        min(line1[0][1], line1[1][1]) < y < max(line1[0][1], line1[1][1]) and \
        min(line2[0][1], line2[1][1]) < y < max(line2[0][1], line2[1][1])


def crossing(a, b):
    return segments_crossing(
        ((a.down.xpos, a.down.height), (a.up.xpos, a.up.height)), 
        ((b.down.xpos, b.down.height), (b.up.xpos, b.up.height))
        )

def get_all_crossing_pairs(lineages):
    crossing_pairs = []
    for i in range(len(lineages)):
        for j in range(i+1, len(lineages)):
            if crossing(lineages[i], lineages[j]):
                crossing_pairs.append([i, j])
    return crossing_pairs

def reduce_crossovers(nodes):
    lineages = get_parent_lineages(nodes)
    lineages = [x for x in lineages if x.up is not None]
    crossing_pairs = get_all_crossing_pairs(lineages)

    for i, j in crossing_pairs:
        # swap downs
        lineages[i].down.xpos, lineages[j].down.xpos = lineages[j].down.xpos, lineages[i].down.xpos
        new_crossing_pairs = get_all_crossing_pairs(lineages)
        if not len(new_crossing_pairs) < len(crossing_pairs):
            # swap downs back
            lineages[i].down.xpos, lineages[j].down.xpos = lineages[j].down.xpos, lineages[i].down.xpos
        else:
            crossing_pairs = new_crossing_pairs

        # swap ups
        lineages[i].up.xpos, lineages[j].up.xpos = lineages[j].up.xpos, lineages[i].up.xpos
        new_crossing_pairs = get_all_crossing_pairs(lineages)
        if not len(new_crossing_pairs) < len(crossing_pairs):
            # swap ups back
            lineages[i].up.xpos, lineages[j].up.xpos = lineages[j].up.xpos, lineages[i].up.xpos
        else:
            crossing_pairs = new_crossing_pairs


        leaf_idx = [i for i, n in enumerate(nodes) if type(n) is Leaf]

        if type(lineages[i].down) is Leaf and type(lineages[j].down) is not Leaf:
            # try to find new position for leaf i

            for k in leaf_idx:
                # swap
                lineages[i].down.xpos, lineages[k].down.xpos = lineages[k].down.xpos, lineages[i].down.xpos

                new_crossing_pairs = get_all_crossing_pairs(lineages)
                if not len(new_crossing_pairs) < len(crossing_pairs):
                    # swap back
                    lineages[i].down.xpos, lineages[k].down.xpos = lineages[k].down.xpos, lineages[i].down.xpos
                else:
                    break


        if type(lineages[i].down) is Leaf and type(lineages[j].down) is not Leaf:
            # try to find new position for leaf j
            for k in leaf_idx:
                # swap
                lineages[j].down.xpos, lineages[k].down.xpos = lineages[k].down.xpos, lineages[j].down.xpos

                new_crossing_pairs = get_all_crossing_pairs(lineages)
                if not len(new_crossing_pairs) < len(crossing_pairs):
                    # swap back
                    lineages[j].down.xpos, lineages[k].down.xpos = lineages[k].down.xpos, lineages[j].down.xpos
                else:
                    break


    # TODO: if cross over involves an internal and an external branch, see if
    # you can move the leaf to a new position that removes the crossover



    # leaf_idx = [i for i, n in enumerate(nodes) if type(n) is Leaf]
    # suffled_leaf_idx = leaf_idx[:]
    # shuffle(suffled_leaf_idx)
    # for i, j in sorted(zip(leaf_idx, suffled_leaf_idx)):
    #     print(i, j)
    #     nodes[i].xpos = nodes[j].xpos

    # new_crossing_pairs = get_all_crossing_pairs(lineages)
    # if not len(new_crossing_pairs) < len(crossing_pairs):
    #     for i, j in sorted(zip(leaf_idx, suffled_leaf_idx), reverse=True):
    #         nodes[i].xpos = nodes[j].xpos




def branch_length(lineage):
    return ((lineage.down.xpos - lineage.up.xpos)**2 + (lineage.down.height - lineage.up.height)**2)**0.5

def reduce_total_branch_length(nodes, shift):
    lineages = get_parent_lineages(nodes)
    lineages = [x for x in lineages if x.up is not None]

    cur_tot = sum(branch_length(l) for l in lineages)
    for i in range(len(lineages)):
        orig_xpos = lineages[i].up.xpos
        lineages[i].up.xpos += shift
        tot = sum(branch_length(l) for l in lineages)
        if tot < cur_tot:
            cur_tot = tot
        else:
            lineages[i].up.xpos = orig_xpos
        lineages[i].up.xpos -= shift
        tot = sum(branch_length(l) for l in lineages)
        if tot < cur_tot:
            cur_tot = tot
        else:
            lineages[i].up.xpos = orig_xpos

def redistribute_leaves(nodes):

    leaves = [n for n in nodes if type(n) is Leaf]
    # min_x = min(n.xpos for n in leaves)
    # max_x = max(n.xpos for n in leaves)
    leaves = sorted(leaves, key=lambda x: x.xpos)
    min_x = 0
    max_x = 1
    for xpos, leaf in zip(np.linspace(min_x, max_x, len(leaves)), leaves):
        leaf.xpos = xpos

####################################################

def add_node_x_positions(nodes):
    """
    Adds x positions in place
    """
    offset = len(nodes)-1
    nodes[-1].xpos = offset
    x_positions_traverse(nodes[-1], offset)

    shift_list = [0.5, 0.4, 0.3, 0.2, 0.1, 0.05, 0.02]

    rescale_positions(nodes)

    [reduce_crossovers(nodes) for _ in range(5)]

    redistribute_leaves(nodes)
    [reduce_crossovers(nodes) for _ in range(5)]
    [reduce_total_branch_length(nodes, shift) for shift in shift_list]
    rescale_positions(nodes)
    [reduce_crossovers(nodes) for _ in range(5)]



    # rescale_positions(nodes)

    # shift_list = [0.5, 0.4, 0.3, 0.2, 0.1, 0.05, 0.02]
    # # shift_list = [0.5, 0.4, 0.3, 0.2]

    # [reduce_crossovers(nodes) for _ in range(5)]

    # # [reduce_total_branch_length(nodes, shift) for shift in shift_list]
    # # [reduce_crossovers(nodes) for _ in range(4)]

    # redistribute_leaves(nodes)

    # [reduce_crossovers(nodes) for _ in range(5)]

    # # [reduce_total_branch_length(nodes, shift) for shift in shift_list]

    # [reduce_total_branch_length(nodes, shift) for shift in shift_list]

    # # [reduce_crossovers(nodes) for _ in range(5)]

    # # redistribute_leaves(nodes)
    # # [reduce_crossovers(nodes) for _ in range(5)]
    # # redistribute_leaves(nodes)
    # # [reduce_crossovers(nodes) for _ in range(5)]


    # rescale_positions(nodes)



#    print('====')

def get_arg_nodes(n=5, N=10000, r=1e-8, L=5e3, simulation="arg"):
    """
    Simulates an ARG
    """
    # because we use the sequence interval from 0 to 1
    r = r * L

#    print('\nSIM:')


    assert simulation in ["arg", "smcprime", "smc"]

    nodes = list()
    live = list() # live lineages

    for i in range(n):
        leaf = Leaf(i, height=0)
        nodes.append(leaf)
        lin = Lineage(lineageid=i, down=leaf, intervals=[(0, 1)])
        leaf.parent = lin
        live.append(lin)
        last_node = i # max node number used so far
        last_lineage = i
        
    while len(live) > 1:
        shuffle(live)

        coal_prob = (len(live) * len(live)-1) / 2 / (2*N) 
        tot_ancestral = sum(interval_sum(x.intervals) for x in live)
        rec_prob = r * tot_ancestral
        wating_time = exponential(1 / (coal_prob + rec_prob))
        height = wating_time + nodes[-1].height
        if random() < coal_prob / (coal_prob + rec_prob):
            # coalescence

            # if full arg:            
            a, b = 0, 1

            if simulation == "smcprime":
                # intervals must overlap or be adjacent:
                while not (interval_any_shared_borders(live[a].intervals, live[b].intervals) or \
                    interval_intersect(live[a].intervals, live[b].intervals)):
                    shuffle(live)
            elif simulation == "smc":
                # intervals must overlap:
                while not interval_intersect(live[a].intervals, live[b].intervals):
                    shuffle(live)
    
            lin_a, lin_b = live.pop(0), live.pop(0)

            intervals = interval_union(lin_a.intervals, lin_b.intervals)

            # new node
            if lin_a.down is lin_b.down:
                # make sure diamond is symmetric
                children=[lin_a.down.left_parent, lin_b.down.right_parent]
            else:
                children=[lin_a, lin_b]  
            node_c = Coalescent(nodeid=last_node+1, height=height, 
                        children=children)
            
            last_node += 1
            nodes.append(node_c)

            # add node to top of coalescing lineages
            lin_a.up = node_c
            lin_b.up = node_c

            # new lineage
            lin_c = Lineage(lineageid=last_lineage+1, down=node_c, intervals=intervals) # fixme intervals
            last_lineage += 1
            live.append(lin_c)
            node_c.parent = lin_c
        else:
            # recombination

            # rec_lin = choices(live, weights=[interval_sum(x.intervals) for x in live], k=1)[0]
            rec_lin = choices(live, weights=[interval_span(x.intervals) for x in live], k=1)[0]
            live.remove(rec_lin)

            # total ancestral material
            total_anc = interval_sum(rec_lin.intervals)

            # recombination point in ancestral material
            recomb_point_anc = random() * total_anc

            # recombination point in full sequence
            cum = 0
            for s, e in rec_lin.intervals:
                cum += e - s
                if cum > recomb_point_anc:
                    recomb_point = e - (cum - recomb_point_anc)
                    break

            # recombination node
            rec_node = Recombination(nodeid=last_node+1, height=height, 
                child=rec_lin, recomb_point=recomb_point)
            last_node += 1
            nodes.append(rec_node)

            # two new lineages both refering back to recombination node
            intervals_a, intervals_b = interval_split(rec_lin.intervals, recomb_point)
            assert interval_sum(intervals_a) and interval_sum(intervals_b)
            assert sum(e-s for s, e in intervals_a) + sum(e-s for s, e in intervals_b) == sum(e-s for s, e in rec_lin.intervals)
            lin_a = Lineage(lineageid=last_lineage+1, down=rec_node, intervals=intervals_a)
            last_lineage += 1
            lin_b = Lineage(lineageid=last_lineage+1, down=rec_node, intervals=intervals_b)
            last_lineage += 1
            live.append(lin_a)
            live.append(lin_b)

            # add parents of node
            rec_node.left_parent = lin_a
            rec_node.right_parent = lin_b

            # add parent for child
            rec_lin.up = rec_node

    add_node_x_positions(nodes)

    return nodes

def _rescale_layout(pos, scale=1):
    # rescale to (0,pscale) in all axes

    # shift origin to (0,0)
    lim = 0  # max coordinate for all axes
    for i in range(pos.shape[1]):
        pos[:, i] -= pos[:, i].min()
        lim = max(pos[:, i].max(), lim)
    # rescale to (0,scale) in all directions, preserves aspect
    for i in range(pos.shape[1]):
        pos[:, i] *= scale / lim
    return pos

def get_positions(nodes):
    """
    Gets list of x,y positions for nodes
    """
    positions = list()
    for node in nodes:
        positions.append((node.xpos, node.height))
    return positions

def get_breakpoints(nodes):
    """
    Gets list of recombination break points
    """    
    # get list of break points
    breakpoints = list()
    for n in nodes:
        if type(n) is Recombination:
#        if type(n) is Recombination and all(p.up in nodes for p in n.parents):
            breakpoints.append(n.recomb_point)
    return sorted(breakpoints)

def get_parent_lineages(nodes, root=True):
    """
    Get list of parent lineages of nodes ordered by id
    """

    lineages = list()
    for node in nodes:
        if type(node) is Coalescent:
            lineages.append(node.parent)
        if type(node) is Recombination:
            lineages.append(node.left_parent)
            lineages.append(node.right_parent)
        if type(node) is Leaf:
            lineages.append(node.parent)

    # get unique and sort them so lineageid matches index
    lineages = list(set(lineages))
    lineages.sort(key=lambda x: x.lineageid)

    if not root:
        # remove dangling root lineage
        lineages = lineages[:-1]

    return lineages

def get_child_lineages(nodes):
    """
    Get list of child lineages of nodes ordered by id
    """

    lineages = list()
    for node in nodes:
        if type(node) is Coalescent:
            lineages.extend(node.children)
        if type(node) is Recombination:
            lineages.append(node.child)

    # get unique and sort them so lineageid matches index
    lineages = list(set(lineages))
    lineages.sort(key=lambda x: x.lineageid)

    return lineages

# def traverse_marginal(node, interval):
#     """
#     Recursive function for getting marginal tree/ARG
#     """    
#     node = deepcopy(node) # TODO: remove if input is a cloned arg.
#     tree_nodes = set()
#     if type(node) is Leaf:
#         tree_nodes.add(node)
#     if type(node) is Recombination:
#         if interval_intersect([interval], node.child.intervals):
#             tree_nodes.add(node)
#             tree_nodes.update(traverse_marginal(node.child.down, interval))
#     elif type(node) is Coalescent:
#         if node.parent is None or interval_intersect([interval], node.parent.intervals):
#             tree_nodes.add(node)
#         del_child = None
#         for i, child in enumerate(node.children):
#             if interval_intersect([interval], child.intervals):
#                 tree_nodes.update(traverse_marginal(child.down, interval))
#             else:
#                 del_child = i
#         if del_child is not None:
#             del node.children[del_child]
#     return tree_nodes

def _traverse_marginal(node, interval):
    """
    Recursive function for getting marginal tree/ARG
    """    
    tree_nodes = set()
    if type(node) is Leaf:
        tree_nodes.add(node)
    if type(node) is Recombination:
        if interval_intersect([interval], node.child.intervals):
            tree_nodes.add(node)
            tree_nodes.update(_traverse_marginal(node.child.down, interval))
    elif type(node) is Coalescent:
        if node.parent is None or interval_intersect([interval], node.parent.intervals):
            tree_nodes.add(node)
        del_child = None
        for i, child in enumerate(node.children):
            if interval_intersect([interval], child.intervals):
                tree_nodes.update(_traverse_marginal(child.down, interval))
            else:
                del_child = i
        if del_child is not None:
            del node.children[del_child]
    return tree_nodes       

def traverse_marginal(node, interval):
    clone = deepcopy(node) # TODO: remove if input is a cloned arg.
    return _traverse_marginal(clone, interval)


def remove_dangling_root(tree_nodes):
    """
    Remove the nodes of a marginal tree or ARG 
    from root to first coalescence
    """    
    for_del = list()
    for i in range(len(tree_nodes)-1, -1, -1):
        if type(tree_nodes[i]) is Coalescent and len(tree_nodes[i].children) == 2:
            break
        for_del.append(i)
    for i in for_del:
        del tree_nodes[i]

def marginal_arg(nodes, interval, strip_dangling_root=True):
    """
    Gets the marginal ARG given a sequene interval
    """

    # TODO: 
    # nodes = clone_arg(nodes)

    # get nodes for marginal
    marg_nodes = traverse_marginal(nodes[-1], list(interval))

    # set to list
    marg_nodes = list(marg_nodes)

    # sort on height
    marg_nodes.sort(key=lambda x: x.height)
    # prune top path above last coalescence
    if strip_dangling_root:
        remove_dangling_root(marg_nodes)

    return marg_nodes

def marginal_trees(nodes, interval, strip_dangling_root=False):
    """
    Gets list of marginal trees
    """
    tree_list = list()
    breakpoints = get_breakpoints(nodes)

    borders = [0] + breakpoints + [1]
    
    # for interval in zip(borders[:-1], borders[1:]):
    #     marg_nodes = marginal_arg(nodes, interval)
    #     tree_list.append(marg_nodes)
    # return tree_list

    interval_list = list()
    for interv in zip(borders[:-1], borders[1:]):
        marg_nodes = marginal_arg(nodes, interv, strip_dangling_root=strip_dangling_root)
        # TODO: make this correct:...
        if marg_nodes and interv[1] > interval[0] and interval[1] > interv[0]:
            tree_list.append(marg_nodes)
            interval_list.append(interv)
    return tree_list, interval_list

# def draw_graph(nodes):
#     """
#     Draws graph using matplotlib
#     """    
#     # make a graph and positions list
#     positions = list()
#     arg = nx.Graph()
#     for node in nodes:
#         arg.add_node(node.nodeid)
#         positions.append((node.xpos, node.height))
#         if isinstance(node, Recombination):
#             arg.add_edge(node.child.down.nodeid, node.nodeid)
#         elif isinstance(node, Coalescent):
#             for child in node.children:
#                 arg.add_edge(child.down.nodeid, node.nodeid)

#     positions = np.array(positions)
#     positions = dict(zip(arg.nodes(), positions))

#     #pos = nx.spring_layout(arg)
#     nx.draw(arg, positions, alpha=0.5, node_size=200, with_labels=True)
#     #with_labels=True, 
#     #connectionstyle='arc3, rad = 0.1', 
#     #arrowstyle='-')

#     # G = nx.DiGraph() #or G = nx.MultiDiGraph()
#     # G.add_node('A')
#     # G.add_node('B')
#     # G.add_edge('A', 'B', length = 2)
#     # G.add_edge('B', 'A', length = 3)
#     # pos = nx.spring_layout(G)
#     # nx.draw(G, pos, with_labels=True, connectionstyle='arc3, rad = 0.1')
#     # edge_labels=dict([((u,v,),d['length'])
#     #              for u,v,d in G.edges(data=True)])

#     plt.show()

def rescale_positions(nodes):
    """
    Rescales xpos and heights to 0,1 range
    """
    max_height = max(n.height for n in nodes)
    max_xpos = max(n.xpos for n in nodes)
    min_xpos = min(n.xpos for n in nodes)
    for node in nodes:
        node.xpos = (node.xpos - min_xpos) / (max_xpos-min_xpos)
        node.height = node.height / max_height

def arg2json(nodes):

    lineages = get_parent_lineages(nodes)

    json_data = dict(Coalescent=[], Recombination=[], Leaf=[], 
                        Lineage=[l.get_dict() for l in lineages])
    for node in nodes:
        json_data[node.__class__.__name__].append(node.get_dict())

    return json.dumps(json_data, indent=4)

def json2arg(json_str):

    nodes = list()

    data = json.loads(json_str)

    # make lineages (with indexes instead of node refs)
    lineages = [Lineage(**data) for data in data['Lineage']]

    for node_data in data['Coalescent']:            
        # make the node
        node = Coalescent(**node_data)
        # populate the parent and children with actual lineages
        node.parent = lineages[node.parent]
        node.children = [lineages[i] for i in node.children]
        nodes.append(node)

    for node_data in data['Recombination']:
        # make the node
        node = Recombination(**node_data)
        # populate the parents and child with actual lineages
        node.left_parent = lineages[node.left_parent]
        node.right_parent = lineages[node.right_parent]
        node.child = lineages[node.child]
        nodes.append(node)

    for node_data in data['Leaf']:
        # make the node
        node = Leaf(**node_data)
        # populate the parent with actual lineages
        node.parent = lineages[node.parent]
        nodes.append(node)

    nodes.sort(key=lambda x: x.nodeid)

    # populate up and down with the nodes
    for lineage in lineages:
        lineage.down = nodes[lineage.down]
        if lineage.up is not None:
            lineage.up = nodes[lineage.up]

    return nodes

if __name__ == '__main__':

    # get arg and add positions
    nodes = get_arg_nodes()

    print(nodes)

    print(deepcopy(nodes))

    sys.exit()

    json_str = arg2json(nodes)

    retrieved_nodes = json2arg(json_str)

    print(nodes)
    print(retrieved_nodes)
    print(nodes == retrieved_nodes)



    # get breakpoints
    breakpoints = get_breakpoints(nodes)
    print(breakpoints)

    # get marginal trees
    trees = marginal_trees(nodes)
    print(trees)

    # marginal arg for some consequtive intervals
    marg_arg = marginal_arg(nodes, [0, breakpoints[1]])


    # # draw graphs for testing
    # draw_graph(nodes)
    # draw_graph(marg_arg)

    # draw_graph(nodes)
    # for tree in marginal_trees(nodes):
    #     print([n.xpos for n in tree])
    #     draw_graph(tree)




import logging
logging.basicConfig(level=logging.INFO)

from ast import Import
import plotly.graph_objects as go

import json
from textwrap import dedent as d

import dash
from dash import dcc
from dash import html
from dash.dependencies import Input, Output, State

import dash_bootstrap_components as dbc

import math
import sys

import logging

#from app import app

# external_stylesheets = ['https://codepen.io/chriddyp/pen/bWLwgP.css']
# external_stylesheets = [dbc.themes.GRID]
external_stylesheets = [dbc.themes.BOOTSTRAP]

try:
    from jupyter_dash import JupyterDash
except ImportError:
    print('jupyter_dash required: install using "conda install -c plotly conda install -c plotly jupyter-dash"')
    sys.exit()
    
app = JupyterDash(__name__, external_stylesheets=external_stylesheets)
# app = dash.Dash(__name__, external_stylesheets=external_stylesheets)

server = app.server
app.config.suppress_callback_exceptions = True

import json

###########################################e

import pandas as pd
import networkx as nx

import arg
from arg import Coalescent, Recombination, Leaf, interval_sum, interval_diff, interval_intersect, get_breakpoints, get_child_lineages, rescale_positions, marginal_arg, traverse_marginal, marginal_trees

import plotly.colors

def get_continuous_color(colorscale, intermed):
    """
    Plotly continuous colorscales assign colors to the range [0, 1]. This function computes the intermediate
    color for any value in that range.

    Plotly doesn't make the colorscales directly accessible in a common format.
    Some are ready to use:
    
        colorscale = plotly.colors.PLOTLY_SCALES["Greens"]

    Others are just swatches that need to be constructed into a colorscale:

        viridis_colors, scale = plotly.colors.convert_colors_to_same_type(plotly.colors.sequential.Viridis)
        colorscale = plotly.colors.make_colorscale(viridis_colors, scale=scale)

    :param colorscale: A plotly continuous colorscale defined with RGB string colors.
    :param intermed: value in the range [0, 1]
    :return: color in rgb string format
    :rtype: str
    """
    if len(colorscale) < 1:
        raise ValueError("colorscale must have at least one color")

    if intermed <= 0 or len(colorscale) == 1:
        return colorscale[0][1]
    if intermed >= 1:
        return colorscale[-1][1]

    for cutoff, color in colorscale:
        if intermed > cutoff:
            low_cutoff, low_color = cutoff, color
        else:
            high_cutoff, high_color = cutoff, color
            break

    # noinspection PyUnboundLocalVariable
    return plotly.colors.find_intermediate_color(
        lowcolor=low_color, highcolor=high_color,
        intermed=((intermed - low_cutoff) / (high_cutoff - low_cutoff)),
        colortype="rgb")

layout = html.Div(
    [
        # Hidden div inside the app that stores the intermediate value
        html.Div(id='intermediate-value', style={'display': 'none'}),

        # row for arg and marginal trees
        dbc.Row(
            [
                # column for arg 
                dbc.Col(
                    [
                        # arg
                        dbc.Container(
                            [
                                dbc.Container(
                                    [

                                        dbc.Row(
                                            [
                                                dbc.Col(
                                                    [
                                                        html.B("Simulation:"),
                                                        # "Simulation:",
                                                        dcc.Dropdown(
                                                            id='sim-dropdown',
                                                            options=[
                                                                {'label': "ARG", 'value': 'arg'},
                                                                # {'label': "SMC", 'value': 'smc'},
                                                                {'label': "SMC'", 'value': 'smcprime'},
                                                                {'label': "SMC", 'value': 'smc'}
                                                            ],
                                                            value='arg', searchable=False, clearable=False
                                                        ),
                                                    ], width=2
                                                ),                                                        
                                                dbc.Col(
                                                    [
                                                        html.B("Nr samples:"),
                                                        dcc.Dropdown(
                                                            id='samples-dropdown',
                                                            options=[
                                                                {'label': "3", 'value': 3},
                                                                {'label': "4", 'value': 4},
                                                                {'label': "5", 'value': 5}
                                                            ],
                                                            value=5, searchable=False, clearable=False,
                                                            style={
                                                                    # 'height': '20px', 
                                                                    # 'width': '80px', 
                                                                    'font-size': "0.85rem",
                                                                    # 'min-height': '1px',
                                                                    },
                                                        ),
                                                    ], width=2
                                                ),  
                                                dbc.Col(
                                                    [
                                                        html.B("Length:"),
                                                        dcc.Dropdown(
                                                            id='seqlen-dropdown',
                                                            options=[
                                                                {'label': "1kb", 'value': 1e+3},
                                                                {'label': "2kb", 'value': 2e+3},
                                                                {'label': "4kb", 'value': 4e+3}
                                                            ],
                                                            value=2e+3, searchable=False, clearable=False,
                                                            style={
                                                                    # 'height': '20px', 
                                                                    # 'width': '80px', 
                                                                    'font-size': "0.85rem",
                                                                    # 'min-height': '1px',
                                                                    },
                                                        ),
                                                    ], width=2
                                                ),                                                                                                        
                                                dbc.Col(
                                                    [ 
                                                        # html.Button('New simulation', id='new-arg-button')
                                                        dbc.Button('New', id='new-arg-button', 
                                                            color="primary", #size="sm", #outline=True,
                                                            style={'height': 35, 'font-size': "0.85rem"},
                                                            className="mr-1"
                                                            )
                                                    ], width=3
                                                ),    
                                                dbc.Col(
                                                    [
                                                        html.Div(id='arg-header'),

                                                        # dcc.Markdown(d("""
                                                        # **Ancestral recombination graph:**   
                                                        # Nodes are colored by amount of ancestral sequence.
                                                        # """), ),                    
                                                    ], width=3
                                                ),                                                                                                                                                                                                        
                                            ], justify="between", align="end", style={'padding': 3} 
                                        ),
                                        dcc.Graph(id='arg-figure',
                                                clear_on_unhover=True,
                                                figure={'layout': {
                                                            'height': 570,
                                                            # 'margin': {'l': 0, 'b': 0, 't': 0, 'r': 0},
                                                                }
                                                            },
                                                    ),
                                    ], className='pretty_container', fluid=True,
                                ),
                            ], style={'padding': 20}
                        ),
                    ], width=8, 
                ),

                # column for marginal trees
                dbc.Col(
                    [
                        dbc.Container(
                            [
                                dbc.Container(
                                    [
                                        dcc.Markdown(d("""
                                        **Marginal tree(s):** Hover over an ARG node.
                                        """), ),                    
                                        dcc.Graph(id='marginal-tree',
                                                    figure={'layout': {
                                                            # 'title': 'Marginal tree',
                                                            'height': 250,
                                                            # 'margin': {'l': 10, 'b': 0, 't': 0, 'r': 0},
                                                                }
                                                            },),
                                    ], className='pretty_container'
                                ),
                            ], style={'padding': 20, 'padding-left': 0, 'padding-bottom': 0}
                        ),
                        dbc.Container(
                            [
                                dbc.Container(
                                    [
                                        dcc.Markdown(d("""
                                        **Ancestral sequences:** Hover over an ARG node.
                                        """), ),                    
                                        dcc.Graph(id='ancestral-sequence',
                                                figure={'layout': {
                                                    'height': 250,
                                                        }
                                                        },),
                                    ], className='pretty_container'
                                ),            
                            ], style={'padding': 20, 'padding-left': 0}
                        ),            
                    ], width=4, 
                ),        
            ], 
            className="g-0"
        ),

        dbc.Row(
            [
                dbc.Col(
                    [
                        dbc.Container(
                            [
                                dbc.Container(
                                    [
                                        dbc.Container(
                                            [
                                                dcc.Markdown(d("""
                                                **Coalesce and recombination events:** 
                                                Slide to see progression of events.
                                                """)),
                                            ], 
                                        ),
                                        dbc.Container(
                                            [
                                                dcc.Slider(
                                                    id='event-slider',
                                                    min=0, max=40, value=0, 
                                                    marks={str(i): str(i) for i in range(0, 40)}, 
                                                    step=None,
                                                    ),
                                            ], style={'padding-bottom': 20}
                                        )
                                    ], className='pretty_container'
                                )
                            ], style={'padding': 20, 'padding-top': 0},
                        )
                    ], width=6
                ),  
                dbc.Col(
                    [
                        dbc.Container(
                            [
                                dbc.Container(
                                    [
                                        dbc.Container(
                                            [
                                                dcc.Markdown(d("""
                                                    **Recombination points:** 
                                                    Slide to see graph for only part of the sequence.
                                                """)),
                                            ]
                                        ),
                                        dbc.Container(
                                            [
                                                dcc.RangeSlider(
                                                    id='seq-slider',
                                                    min=0,
                                                    max=1000,
                                                    value=[0, 1000],
                                                    # step=None,
                                                    marks={0: '0', 1000: '1'},
                                                    pushable=30,
                                                )
                                            ], style={'padding-bottom': 20}
                                        ),
                                    ], className='pretty_container', 
                                ),
                            ], style={'padding': 20, 'padding-left': 0, 'padding-top': 0}
                        ),
                    ], width=6, align='start',
                ),
            ],
            className="g-0"
        ),
    ], style={'padding': 20}
)


def get_bezier_points(x1, y1, x2, y2, offset=1, absolute=True):
    mid_x = x1 + (x2 - x1) / 2
    mid_y = y1 + (y2 - y1) / 2
    length = math.sqrt((x2-x1)**2 + (y2-y1)**2)
    if absolute:
        hyp = offset
    else:
        hyp = length / offset 
        
    if y2 == y1:
        b11, b12 = mid_x, mid_y + hyp
        b21, b22 = mid_x, mid_y - hyp    
    elif x2 == x1:
        b11, b12 = mid_x + hyp, mid_y
        b21, b22 = mid_x - hyp, mid_y
    else:
        slope = (y2-y1) / (x2-x1)
        if slope < 0:
            angle_in_radians = math.atan(1/slope)
        else:
            angle_in_radians = math.atan(slope)   
        kat = math.sin(angle_in_radians) * hyp
        b11, b12 = mid_x + kat*slope, mid_y + kat/slope
        b21, b22 = mid_x - kat*slope, mid_y - kat/slope
    
    return b11, b12, b21, b22


def arg_figure_data(nodes):

    traces = []

    edge_x = []
    edge_y = []
    
    diamond_shapes = []
    
    # for lineage in get_parent_lineages(nodes, root=False):
    for lineage in get_child_lineages(nodes):
        
        if type(lineage.down) is Recombination and \
            type(lineage.up) is Coalescent and \
            set(lineage.up.children) == set([lineage.down.right_parent, lineage.down.left_parent]):

            # diamond recombination:
            x1 = lineage.down.xpos
            y1 = lineage.down.height

            x2 = lineage.up.xpos
            y2 = lineage.up.height
            
            b11, b12, b21, b22 = get_bezier_points(x1, y1, x2, y2, offset=0.02)

            diamond_shapes.append(
                dict(
                    type="path",
                    path=f"M {x1},{y1} Q {b11},{b12} {x2},{y2}",
                    # line_color="lightgray",
                    layer='below',
                    line= {'width': 2, 'color': 'gray'}
                )
            )
            diamond_shapes.append(
                dict(
                    type="path",
                    path=f"M {x1},{y1} Q {b21},{b22} {x2},{y2}",
                    # line_color="lightgray",
                    layer='below',
                    line= {'width': 2, 'color': 'gray'},
                )
            )            

        else:        
            # start
            edge_x.append(lineage.down.xpos)
            edge_y.append(lineage.down.height)
            # end
            edge_x.append(lineage.up.xpos)
            edge_y.append(lineage.up.height)
            # gap
            edge_x.append(None)
            edge_y.append(None)

    traces.append(dict(
        x=edge_x,
        y=edge_y,
        mode='lines',
        opacity=1,
        hoverinfo = 'skip',
        line={
            'color': 'grey',
        },
        name=''
    ))

    node_x = []
    node_y = []    
    node_text = []
    node_color = []
    for node in nodes:
        node_x.append(node.xpos)
        node_y.append(node.height)
        prop_ancestral = 1
        if type(node) is Coalescent:
            prop_ancestral = interval_sum(node.parent.intervals)
        elif type(node) is Recombination:
            prop_ancestral = interval_sum(node.child.intervals)
        node_text.append(f"Fraction ancestral: {round(prop_ancestral, 2)}<br>Event: {type(node).__name__}")

        node_color.append(prop_ancestral)

    traces.append(dict(
        x=node_x,
        y=node_y,
        text=node_text,
        # range_color=[0, 1],
        # cmin=0,
        # cmax=1,
        mode='markers',
        opacity=1,
        hoverinfo ='text',
        marker={
            'size': 10,
            'color': node_color,
            'cmin': 0,
            'cmax': 1,
            'line': {'width': 0.7, 'color': 'white'},
            # 'colorscale': 'Viridis',
            'colorscale': 'Rainbow',
            'colorbar': {'title': 'Fraction<br>ancestral<br>sequence',
                        'titleside': 'top',
                        'thickness': 15,
                        'len': 0.5,
                        # 'tickmode': 'array',
                        'tickvals': [0, 0.5, 1],
                        # 'ticktext': ['0', '1'],
                        'ticks': 'outside',
                        },
        },
        name=''
    ))

    return dict(data=traces,
                layout=dict(xaxis=dict(fixedrange=True, 
                                       range=[-0.1, 1.1], #title='Samples',
                                       showgrid=False, showline=False, 
                                       zeroline=False, showticklabels=False
                                       ),
                            yaxis=dict(fixedrange=True, 
                                       range=[-0.1, 1.1], #title='Time',
                                       showgrid=False, showline=False, 
                                       zeroline=False, showticklabels=False
                                       ),
                            hovermode='closest',
                            range_color=[0,1],
                            margin= {'l': 50, 'b': 20, 't': 20, 'r': 20},
                            transition = {'duration': 0},
                            showlegend=False,
                            shapes=diamond_shapes,
                            )
                )


def tree_figure_data(node_lists):

    traces = []

    edge_x = []
    edge_y = [] 
    node_x = []
    node_y = []    
    node_color = []

    for i, nodes in enumerate(node_lists):

        # for lineage in get_parent_lineages(nodes, root=False):
        for lineage in get_child_lineages(nodes):
            # start
            edge_x.append(lineage.down.xpos)
            edge_y.append(lineage.down.height)
            # end
            edge_x.append(lineage.up.xpos)
            edge_y.append(lineage.up.height)
            # gap
            edge_x.append(None)
            edge_y.append(None)

        for node in nodes:
            node_x.append(node.xpos)
            node_y.append(node.height)

            node_color.append(i/len(node_lists))

    traces.append(dict(
        x=edge_x,
        y=edge_y,
        mode='lines',
        opacity=1,
        hoverinfo = 'skip',
        line={
            'color': 'grey',
        },
        name=''
    ))

    traces.append(dict(
        x=node_x,
        y=node_y,
        mode='markers',
        opacity=1,
        hoverinfo ='text',
        marker={
            'size': 7,
            'color': node_color,
            'cmin': 0,
            'cmax': 1,
            'colorscale': 'Rainbow',
            'line': {'width': 0.3, 'color': 'white'},
        },
        name=''
    ))

    return dict(data=traces,
                layout=dict(xaxis=dict(fixedrange=True, 
                                       range=[-0.05, 1.05], #title='Samples',
                                       showgrid=False, showline=False, 
                                       zeroline=False, showticklabels=False
                                       ),
                            yaxis=dict(fixedrange=True, 
                                       range=[-0.1, 1.1], #title='Time',
                                       showgrid=False, showline=False, 
                                       zeroline=False, showticklabels=False
                                       ),
                            hovermode='closest',
                            range_color=[0,1],
                            margin= {'l': 50, 'b': 20, 't': 20, 'r': 20},
                            transition = {'duration': 0},
                            showlegend=False,
                            )
                )


@app.callback(
    Output('arg-header', 'children'),
    [Input('new-arg-button', 'n_clicks')])
def update_header(n_clicks):

    if n_clicks is None:
        n_sim = 1
    else:
        n_sim = n_clicks + 1

    return dcc.Markdown(d("""
                **Simulation #{}:**   
                """.format(n_sim)))

@app.callback(Output('intermediate-value', 'children'), 
    [Input('new-arg-button', 'n_clicks'),
     Input('sim-dropdown', 'value'),
     Input('samples-dropdown', 'value'),
     Input('seqlen-dropdown', 'value')])
def new_data(n_clicks, sim, samples, length):

    nodes = arg.get_arg_nodes(L=length, n=samples, simulation=sim)
#    rescale_positions(nodes)
    json_str = arg.arg2json(nodes)
    return json_str

@app.callback(
    [Output(component_id='event-slider', component_property='min'),
     Output(component_id='event-slider', component_property='max'),
     Output(component_id='event-slider', component_property='step'),
     Output(component_id='event-slider', component_property='value')],
    [Input('intermediate-value', 'children')])    
def update_event_slider(jsonified_data):
    if jsonified_data:
        nodes = arg.json2arg(jsonified_data)
    else:
        nodes = []

    nr_leaves = len([n for n in nodes if type(n) is arg.Leaf])
    nr_events = len(nodes)-nr_leaves
    return 0, nr_events, 1, nr_events

@app.callback(
    [Output(component_id='seq-slider', component_property='min'),
     Output(component_id='seq-slider', component_property='max'),
     Output(component_id='seq-slider', component_property='value'),
     Output(component_id='seq-slider', component_property='marks')],
    [Input('intermediate-value', 'children')])    
def update_seq_slider(jsonified_data):
    if jsonified_data:
        nodes = arg.json2arg(jsonified_data)
    else:
        nodes = []
    breakpoints = get_breakpoints(nodes)
    marks = dict((b*1000, str(i+1)) for i, b in enumerate(breakpoints))
    if not marks:
        marks = None
    return 0, 1000, [0, 1000], marks


@app.callback(
    Output('arg-figure', 'figure'),
    [Input('intermediate-value', 'children'),
     Input('event-slider', 'value'),
     Input('seq-slider', 'value')])
def update_arg_figure(jsonified_data, event, interval):

    if jsonified_data:
        nodes = arg.json2arg(jsonified_data)

        interval = [i/1000 for i in interval]

        # Get marginal arg for interval
        marg_arg_nodes = marginal_arg(nodes, interval)
        # print(interval)
        # get only subset of events
        nr_leaves = len([n for n in nodes if type(n) is arg.Leaf])
        new_nodes = marg_arg_nodes[:nr_leaves+event]
    else:
        new_nodes = []

    return arg_figure_data(new_nodes)


@app.callback(
    Output('marginal-tree', 'figure'),
    [Input('intermediate-value', 'children'),
     Input('arg-figure', 'hoverData'),
     Input('seq-slider', 'value')])
def update_marg_tree_figure(jsonified_data, hover, slider_interval):

    marg_tree_list = []
    if hover and jsonified_data:
        nodes = arg.json2arg(jsonified_data)
        focus_node_idx = hover['points'][0]['pointIndex']
        focus_node = nodes[focus_node_idx]

        if type(focus_node) is Recombination:
            intervals = focus_node.child.intervals
        else:
            intervals = focus_node.parent.intervals

        # slider interval is 0-1000 not 0-1:
        slider_interval = [x/1000 for x in slider_interval]
        # get part of intervals that intersect slider interval:
        intervals = interval_intersect([slider_interval], intervals)

        for interval in intervals:
            # get marginal arg under focus node            
            new_nodes = traverse_marginal(focus_node, interval)
            new_nodes = list(new_nodes)
            new_nodes.sort(key=lambda x: x.height)

            marg_trees, _ = marginal_trees(new_nodes, interval)
            marg_tree_list.extend(marg_trees)
            
        nr_cols = len(marg_tree_list)

        space = 0.5
    
        for i in range(nr_cols):
            tree = marg_tree_list[i]
            for node in tree:
                node.xpos = node.xpos/(nr_cols+(nr_cols-1)*space) + i/nr_cols
            marg_tree_list[i] = tree

    # TODO: Maybe keep "dangling root" branch here

    if marg_tree_list:
        return(tree_figure_data(marg_tree_list))
    else:
        return(tree_figure_data([]))


@app.callback(
    Output('ancestral-sequence', 'figure'),
    [Input('intermediate-value', 'children'),
     Input('arg-figure', 'hoverData'),
     Input('seq-slider', 'value')])    
def update_ancestral_seq_figure(jsonified_data, hover, slider_interval):

    traces = []
    shape_list = []
    
    if hover and jsonified_data:
        nodes = arg.json2arg(jsonified_data)
        focus_node_idx = hover['points'][0]['pointIndex']
        focus_node = nodes[focus_node_idx]

        # slider interval is 0-1000 not 0-1:
        slider_interval = [x/1000 for x in slider_interval]

        gray_segments = list(map(tuple, interval_diff([[0, 1]], [slider_interval])))

        def get_segments(focus_node, intervals):
            segments = list()
            marg_tree_list = list()
            for interval in intervals:  
                new_nodes = traverse_marginal(focus_node, interval)
                new_nodes = list(new_nodes)
                new_nodes.sort(key=lambda x: x.height)
                marg_trees, marg_segm = marginal_trees(new_nodes, interval)
#                 print(marg_trees, marg_segm)
                marg_tree_list.extend(marg_trees)
                segments.extend(marg_segm)
            return segments

        def get_shapes(segments, gray_segments, x, y, color_map):
            shape_list = list()
            shape = dict(type='rect', xref='x', yref='y', fillcolor=None, line= {'width': 1},
                        x0=x, y0=y, x1=x+2/5, y1=y+0.1)
            shape_list.append(shape)                
            for i, segment in enumerate(segments):
                color=color_map[segment]
                shape = dict(type='rect', xref='x', yref='y', fillcolor=color, line= {'width': 1},
                            x0=x+segment[0]*2/5, y0=y, x1=x+segment[1]*2/5, y1=y+0.1)
                shape_list.append(shape)
            for i, segment in enumerate(gray_segments):
                shape = dict(type='rect', xref='x', yref='y', fillcolor='lightgray', line= {'width': 1},
                            x0=x+segment[0]*2/5, y0=y, x1=x+segment[1]*2/5, y1=y+0.1)
                shape_list.append(shape)
            return shape_list

        if type(focus_node) is Leaf:

            colors, _ = plotly.colors.convert_colors_to_same_type(plotly.colors.sequential.Rainbow)
            colorscale = plotly.colors.make_colorscale(colors)
            color = get_continuous_color(colorscale, intermed=0)

            shape_list = [dict(type='rect', xref='x', yref='y', fillcolor=color, line= {'width': 1},
                x0=1.5/5, y0=0.25, x1=3.5/5, y1=0.35)]
            for segment in gray_segments:
                shape = dict(type='rect', xref='x', yref='y', fillcolor='lightgray', line= {'width': 1},
                            x0=1.5/5+segment[0]*2/5, y0=0.25, x1=1.5/5+segment[1]*2/5, y1=0.35)
                shape_list.append(shape)                

        elif type(focus_node) is Recombination:
            # print("###", focus_node.left_parent.intervals, focus_node.right_parent.intervals)
            segments1 = get_segments(focus_node, focus_node.left_parent.intervals)
            segments2 = get_segments(focus_node, focus_node.right_parent.intervals)
            segments3 = get_segments(focus_node, focus_node.child.intervals)

            # get part of intervals that intersect slider interval:
            segments1 = list(map(tuple, interval_intersect([slider_interval], segments1)))
            segments2 = list(map(tuple, interval_intersect([slider_interval], segments2)))
            segments3 = list(map(tuple, interval_intersect([slider_interval], segments3)))

            unique_segments = sorted(set().union(segments1, segments2, segments3))
            color_map = dict()
            colors, _ = plotly.colors.convert_colors_to_same_type(plotly.colors.sequential.Rainbow)
            colorscale = plotly.colors.make_colorscale(colors)
            for i, s in enumerate(unique_segments):
                color_map[s] = get_continuous_color(colorscale, intermed=i/len(unique_segments))

            shape_list = \
                get_shapes(segments1, gray_segments, x=0, y=0.75, color_map=color_map) + \
                get_shapes(segments2, gray_segments, x=3/5, y=0.75, color_map=color_map) + \
                get_shapes(segments3, gray_segments, x=1.5/5, y=0.25, color_map=color_map) + \
                   [dict(type='line', xref='x', yref='y', line= {'width': 2, 'color': 'gray'},
                            x0=0.5, y0=0.55, x1=0.5, y1=0.35),
                    dict(type='line', xref='x', yref='y', line= {'width': 2, 'color': 'gray'},
                            x0=0.5, y0=0.55, x1=1/5, y1=0.75),                            
                    dict(type='line', xref='x', yref='y', line= {'width': 2, 'color': 'gray'},
                            x0=0.5, y0=0.55, x1=4/5, y1=0.75)]

            # print("slider", slider_interval)
            # shape_list.append(dict(type='rect', xref='x', yref='y', fillcolor='grey', line= {'width': 1},
            #             x0=slider_interval[0], y0=0.5, x1=slider_interval[1], y1=0.5+0.1))
                
        else:
            segments1 = get_segments(focus_node, focus_node.children[0].intervals)
            segments2 = get_segments(focus_node, focus_node.children[1].intervals)
            segments3 = get_segments(focus_node, focus_node.parent.intervals)

            # get part of intervals that intersect slider interval:
            segments1 = list(map(tuple, interval_intersect([slider_interval], segments1)))
            segments2 = list(map(tuple, interval_intersect([slider_interval], segments2)))
            segments3 = list(map(tuple, interval_intersect([slider_interval], segments3)))

            unique_segments = sorted(set().union(segments1, segments2, segments3))
            color_map = dict()
            colors, _ = plotly.colors.convert_colors_to_same_type(plotly.colors.sequential.Rainbow)
            colorscale = plotly.colors.make_colorscale(colors)
            for i, s in enumerate(unique_segments):
                color_map[s] = get_continuous_color(colorscale, intermed=i/len(unique_segments))

            shape_list = \
                get_shapes(segments1, gray_segments, x=0, y=0.25, color_map=color_map) + \
                get_shapes(segments2, gray_segments, x=3/5, y=0.25, color_map=color_map) + \
                get_shapes(segments3, gray_segments, x=1.5/5, y=0.75, color_map=color_map) + \
                    [dict(type='line', xref='x', yref='y', line= {'width': 2, 'color': 'gray'},
                            x0=0.5, y0=0.55, x1=0.5, y1=0.75),
                    dict(type='line', xref='x', yref='y', line= {'width': 2, 'color': 'gray'},
                            x0=0.5, y0=0.55, x1=1/5, y1=0.35),                            
                    dict(type='line', xref='x', yref='y', line= {'width': 2, 'color': 'gray'},
                            x0=0.5, y0=0.55, x1=4/5, y1=0.35)]

    figure_data = dict(
                    data=traces,
                    layout=dict(xaxis=dict(fixedrange=True,
                                           range=[-0.01, 1.01], #title='Samples',
                                           showgrid=False, showline=False, 
                                           zeroline=False, showticklabels=False
                                           ),
                               yaxis=dict(fixedrange=True,
                                           range=[0, 1], #title='Time',
                                           showgrid=False, showline=False, 
                                           zeroline=False, showticklabels=False
                                           ),
                               margin= {'l': 0, 'b': 0, 't': 20, 'r': 0},
                               transition = {'duration': 0},
                               showlegend=False,
                               shapes=shape_list,
                               )
                    )

#        figure_data['layout']['shapes'].extend(shape_list)

    return figure_data
    


app.layout = layout

if __name__ == '__main__':
    app.run_server(debug=True)


import setuptools, os, sys

# read the contents of your README file
from os import path
this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

# SEARCH/REPLACE package_name WITH THE ACTUAL PACKAGE NAME BELOW:

setuptools.setup(
    name='popgen-dashboards',
    version="1.0.10",
    author="Kasper Munch",
    author_email="kaspermunch@birc.au.dk",
    description="Short description",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='https://github.com/kaspermunch/popgen_dashboards',
    include_package_data=True,  
    packages=setuptools.find_packages(),
    python_requires='>=3.6',
    install_requires=[
        'jupyterlab',
        'jupyter-dash',
        'dash',
        'dash-bootstrap-components',
        'pandas',
        'networkx'
    ])
